let make_arr dim_x dim_y dim_z =
  let arr = Array.make dim_z (Array.make 0 (Array.make 0 ' ')) in
  for i = 0 to dim_z - 1 do
    arr.(i) <- Array.make dim_x (Array.make 0 ' ');
    for j = 0 to dim_x - 1 do
      arr.(i).(j) <- Array.make dim_y ' ';
    done
  done;
  arr

let array3_dim arr =
  (Array.length arr, Array.length arr.(0), Array.length arr.(0).(0))

let is_on_grid arr (ply, col, row) =
  let (plies, cols, rows) = array3_dim arr in
  not (row < 0 || row >= rows || col < 0 || col >= cols || ply < 0 || ply >= plies)
  
let is_cell arr (ply, row, col) =
  if not @@ is_on_grid arr (ply, row, col) then false
  else arr.(ply).(row).(col) != ' '

let set arr (ply, row, col) char =
  if is_on_grid arr (ply, row, col) then
    arr.(ply).(row).(col) <- char

let ( ++ ) (a, b, c) (x, y, z) = (a + x, b + y, c + z)

let num_cols = 100
                               
let all_deltas col =
  let col_corr = float_of_int col /. float_of_int num_cols *. 1.0 in
  [((0, 0, 1), 0.4);
   ((0, 0, -1), 0.2);
   ((0, 1, 0), 0.1 +. col_corr);
   ((0, -1, 0), 0.1 +. col_corr);
   ((1, 0, 0), 0.03);
   ((-1, 0, 0), 0.03)]

let all_neighbors arr ((_, _, col) as pos) =
  List.filter (fun (v, p) -> is_on_grid arr v)
    (List.map (fun (v, p) -> (v ++ pos, p)) (all_deltas col))

let random_neighbor arr pos =
  let ls = all_neighbors arr pos in
  let sum_prob = List.fold_left ( +. ) 0.0 (List.map snd ls) in
  let r = Random.float sum_prob in
  let rec loop partial_sum = function
    | [] -> assert false
    | [(v, _)] -> v
    | (v, p) :: rest ->
       let higher_sum = partial_sum +. p in
       if higher_sum > r then v
       else loop higher_sum rest in
  loop 0.0 ls

let rec draw_path arr pos counter =
  if counter > 0 then
    begin
      let new_pos = random_neighbor arr pos in
      set arr new_pos '+';
      draw_path arr new_pos (counter - 1)
    end

let place_eggs_boosters arr =
  let (plies, rows, cols) = array3_dim arr in
  for ply = 0 to plies - 1 do
    for row = 0 to rows - 1 do
      for col = 0 to cols - 1 do
        if is_cell arr (ply, row, col)
           && not (is_cell arr (ply - 1, row, col)) then
          let f delta char =
            if not (is_cell arr ((ply, row, col) ++ delta))
               && Random.int 256 < 16 then
              set arr (ply, row, col) char in
          begin
            f (0, 1, 0) 'v';
            f (0, -1, 0) '^';
            f (0, 0, 1) '>';
            f (0, 0, -1) '<';
            if Random.int 256 < 3 then
              set arr (ply, row, col) 'B';
        end
      done
    done
  done
  
let main () =
  Random.init 4;
  let arr = make_arr 20 200 2 in
  draw_path arr (0, 10, 0) 1200;
  place_eggs_boosters arr;
  let stream = open_out "voxelstorm12.vox" in
  output_string stream "#ceiling_tex shared_pk02/ceiling03 0.25 0.25 0 0
#floor_tex shared_pk02/floor01 0.25 0.25 0 0
#wall_tex_ladder shared_pk02/wall07a 0.5 0.5 0 0
#wall_tex_ply 0 shared_pk02/wall07a 0.5 0.5 0 0
#wall_tex_ply 1 shared_pk02/wall07b 0.5 0.5 0 0
#sky_tex voxel/sky 1 1 0 0
#lamp_tex voxel/light2_white 0.5 0.5 64 64
#ladder_width 128
#single_sky
";
  let (plies, rows, cols) = array3_dim arr in
  for ply = 0 to plies - 1 do
    output_string stream "#ply\n";
    for row = 0 to rows - 1 do
      for col = 0 to cols - 1 do
        output_char stream arr.(ply).(row).(col)
      done;
      output_char stream '\n';
    done;
  done;
  close_out stream

let () = main ()
